/* xcam -- X-based camera frontend
   Uses the SANE library.
   Copyright (C) 1997 David Mosberger and Tristan Tarrant

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>

#include <gtk.h>
#include <gtkglue.h>
#include <sane.h>

#define NELEMS(a)	((int) sizeof(a)/sizeof(a[0]))

typedef struct Canvas
  {
    GtkWidget *preview;
    GdkGC *gc;
    GdkImage *gdk_image;
    GdkColormap *graylevel_cmap;	/* for 8 bit displays */
    guint32 graylevel[64];		/* 64 graylevel pixels */
    GdkColormap *cube_cmap;
    GdkColor cube_colors[5 * 6 * 5];
  }
Canvas;

static const char *prog_name;
static const char *default_device;
static GSGDialog *device_dialog;

static struct {
  GtkWidget *shell;
  GtkWidget *device_dialog_box;
  GtkWidget *menubar;
  GtkWidget *play_stop_label;
  struct {
    GtkWidget *item;	/* the menu bar item */
    GtkWidget *menu;	/* the associated menu */
  } devices;
  Canvas canvas;
  gint gdk_input_tag;	/* tag returned by gdk_input_add () */
  int playing;		/* are we playing video? */

  SANE_Byte buf[32768];
  SANE_Int remaining;
  SANE_Parameters params;
  gpointer data;	/* image data */
  int x;		/* x position */
} win;

/* forward declarations: */
static void rescan_devices (GtkWidget *widget, gpointer client_data,
			    gpointer call_data);
static void next_frame (void);


#define CANVAS_EVENT_MASK	GDK_BUTTON1_MOTION_MASK |	\
				GDK_EXPOSURE_MASK |		\
				GDK_BUTTON_PRESS_MASK |		\
				GDK_ENTER_NOTIFY_MASK

static void
display_image (Canvas *canvas)
{
  if (canvas->gdk_image)
    {
      gdk_draw_image (canvas->preview->window, canvas->gc, canvas->gdk_image,
		      0, 0, 0, 0,
		      canvas->gdk_image->width, canvas->gdk_image->height);
      gdk_flush ();
    }
}

static gint
canvas_events (GtkWidget *widget, GdkEvent *event)
{
  Canvas *canvas = (Canvas *) gtk_widget_get_user_data (widget);
  if (!canvas)
    return FALSE;

  switch (event->type)
    {
    case GDK_EXPOSE:
      printf ("canvas_events: expose\n");
      if (!canvas->gc)
	canvas->gc = gdk_gc_new (canvas->preview->window);
      display_image (canvas);
      break;

    case GDK_BUTTON_PRESS:
      printf ("canvas_events: button press\n");
      break;

    case GDK_BUTTON_RELEASE:
      printf ("canvas_events: button release\n");
      break;

    case GDK_MOTION_NOTIFY:
      printf ("canvas_events: motion notify\n");
      break;

    case GDK_ENTER_NOTIFY:
#if 0
      gdk_colors_store (win.canvas.cube_cmap, win.canvas.cube_colors,
			NELEMS(win.canvas.cube_colors));
#endif
      printf ("canvas_events: enter notify\n");
      break;

    default:
      break;
    }
  return FALSE;
}

static void
switch_device (GtkWidget *widget, gpointer client_data, gpointer call_data)
{
  SANE_Int val;

  win.playing = FALSE;

  if (device_dialog)
    gsg_destroy_dialog (device_dialog);

  device_dialog = gsg_create_dialog (win.device_dialog_box, client_data);
}

static void
exit_callback (GtkWidget *widget, gpointer client_data, gpointer call_data)
{
  gsg_destroy_dialog (device_dialog);
  exit (0);
}

static GtkWidget *
build_files_menu (void)
{
  GtkWidget *menu, *item;

  menu = gtk_menu_new ();

  item = gtk_menu_item_new_with_label ("Rescan devices");
  gtk_callback_add (gtk_menu_item_get_state (item), rescan_devices, 0);
  gtk_container_add (menu, item);
  gtk_widget_show (item);

  item = gtk_menu_item_new ();
  gtk_container_add (menu, item);
  gtk_widget_show (item);

  item = gtk_menu_item_new_with_label ("Exit");
  gtk_container_add (menu, item);
  gtk_callback_add (gtk_menu_item_get_state (item), exit_callback, 0);
  gtk_widget_show (item);

  return menu;
}

static GtkWidget *
build_device_menu (void)
{
  const SANE_Device ** device;
  GtkWidget *menu, *item;
  SANE_Status result;
  int i;

  menu = gtk_menu_new ();

  result = sane_get_devices (&device, SANE_FALSE);
  if (result != SANE_STATUS_GOOD)
    {
      fprintf (stderr, "%s: %s\n", prog_name, sane_strstatus (result));
      exit (1);
    }
	
  default_device = 0;
  for (i = 0; device[i]; ++i)
    {
      if (!default_device)
	default_device = device[i]->name;

      item = gtk_menu_item_new_with_label ((char *) device[i]->name);
      gtk_container_add (menu, item);
      gtk_callback_add (gtk_menu_item_get_state (item),
			switch_device, (char *) device[i]->name);
      gtk_widget_show (item);
    }
  return menu;
}

static void
rescan_devices (GtkWidget *widget, gpointer client_data, gpointer call_data)
{
  GtkWidget *old_menu = win.devices.menu;
  win.devices.menu = build_device_menu ();
  gtk_menu_item_set_submenu (win.devices.item, win.devices.menu);

  gtk_widget_destroy (old_menu);
}

static void
input_available (gpointer ignore, gint source, GdkInputCondition cond)
{
  size_t buf_size, remaining = win.remaining;
  int i, x, pixels_per_line, bytes_per_line;
  u_int red, green, blue, lum;
  SANE_Byte * buf = win.buf;
  u_char *data = win.data;
  SANE_Status result;
  SANE_Int len;
  int16_t rgb;

  if (!win.playing)
    {
      /* looks like we got cancelled */
      sane_cancel (device_dialog->dev);
      if (win.gdk_input_tag >= 0)
	gdk_input_remove (win.gdk_input_tag);
      win.gdk_input_tag = -1;
      gtk_label_set (win.play_stop_label, "Play");
      return;
    }

  buf_size = sizeof (win.buf);
  pixels_per_line = win.params.pixels_per_line;
  bytes_per_line = win.canvas.gdk_image->bpl;
  x = win.x;

  while (1)
    {
      result = sane_read (device_dialog->dev,
			  buf + remaining, buf_size - remaining, &len);
      if (result != SANE_STATUS_GOOD)
	{
	  sane_cancel (device_dialog->dev);
	  if (win.gdk_input_tag >= 0)
	    gdk_input_remove (win.gdk_input_tag);
	  else
	    win.playing = FALSE;
	  win.gdk_input_tag = -1;

	  if (result == SANE_STATUS_EOF)
	    {
	      display_image (&win.canvas);
	      if (win.playing)
		next_frame ();			/* arrange for next frame */
	      else
		gtk_label_set (win.play_stop_label, "Play");
	    }
	  else
	    {
	      sprintf (buf, "Error during read: %s.", sane_strstatus (result));
	      gsg_error (buf);
	    }
	  return;
	}

      len += remaining;
      switch (win.params.format)
	{
	case SANE_FRAME_GRAY:
	  for (i = 0; i < len; ++i)
	    *data++ = buf[i];
	  break;

	case SANE_FRAME_RGB:
	  switch (win.canvas.gdk_image->depth)
	    {
	    case 8:	/* XXX should dither */
	      for (i = 0; i <= len - 3; i += 3)
		{
		  red = buf[i + 0]; green = buf[i + 1]; blue = buf[i + 2];
#if 1
		  lum = (3*red + 5*green + 2*blue) / 40;
		  if (lum >= NELEMS(win.canvas.graylevel))
		    lum = NELEMS(win.canvas.graylevel) - 1;
		  *data++ = win.canvas.graylevel[lum];
#else
		  *data++ = 30*(red / 64) + 5*(green / 43) + (blue / 64);
#endif
		  if (++x >= pixels_per_line)
		    {
		      x = 0;
		      data += bytes_per_line - pixels_per_line;
		    }
		}
	      break;

	    case 15:
	      /* what about byte-order? */
	      for (i = 0; i <= len - 3; i += 3)
		{
		  rgb = (  ((buf[i + 0] >> 3) << 10)	/* 5 bits of red */
			 | ((buf[i + 1] >> 3) <<  5)	/* 5 bits of green */
			 | ((buf[i + 2] >> 3) <<  0));	/* 5 bits of blue */
		  ((int16_t *) data) = rgb;
		  data += 2;
		  if (++x >= pixels_per_line)
		    {
		      x = 0;
		      data += bytes_per_line - 2*pixels_per_line;
		    }
		}
	      break;

	    case 16:
	      /* what about byte-order? */
	      for (i = 0; i <= len - 3; i += 3)
		{
		  rgb = (  ((buf[i + 0] >> 3) << 11)	/* 5 bits of red */
			 | ((buf[i + 1] >> 2) <<  5)	/* 6 bits of green */
			 | ((buf[i + 2] >> 3) <<  0));	/* 5 bits of blue */
		  ((int16_t *) data) = rgb;
		  data += 2;
		  if (++x >= pixels_per_line)
		    {
		      x = 0;
		      data += bytes_per_line - 2*pixels_per_line;
		    }
		}
	      break;

	    case 24:
	      for (i = 0; i <= len - 3; i += 3)
		{
		  red = buf[i + 0]; green = buf[i + 1]; blue  = buf[i + 2];
		  data[0] = red; data[1] = green; data[2] = blue;
		  data += 3;
		  if (++x >= pixels_per_line)
		    {
		      x = 0;
		      data += bytes_per_line - 3*pixels_per_line;
		    }
		}
	      break;

	    case 32:
	      for (i = 0; i <= len - 3; i += 3)
		{
		  red = buf[i + 0]; green = buf[i + 1]; blue  = buf[i + 2];
		  data[0] = red; data[1] = green; data[2] = blue; data[3] = 0;
		  data += 4;
		  if (++x >= pixels_per_line)
		    {
		      x = 0;
		      data += bytes_per_line - 4*pixels_per_line;
		    }
		}
	      break;

	    default:
	      i = 0;
	      fprintf (stderr, "input_available: depth %d unexpected!\n",
		       win.canvas.gdk_image->depth);
	      break;
	    }
	  remaining = (len - i);
	  if (remaining > 0)
	    memcpy (buf, buf + i, remaining);
	  break;

	case SANE_FRAME_RED:
	case SANE_FRAME_GREEN:
	case SANE_FRAME_BLUE:
	  fprintf (stderr, "Don't know yet how to deal with R/G/B frames\n");
	  return;
	}
    }

  win.data = data;
  win.x = x;
  win.remaining = remaining;
}

static void
next_frame (void)
{
  char buf[256];
  SANE_Status result;
  int fd;

  result = sane_start (device_dialog->dev);
  if (result != SANE_STATUS_GOOD)
    {
      sprintf (buf, "Failed to start scanner: %s.", sane_strstatus (result));
      gsg_error (buf);
      return;
    }

  result = sane_get_parameters (device_dialog->dev, &win.params);
  if (result != SANE_STATUS_GOOD)
    {
      sprintf (buf, "Failed to get parameters: %s.", sane_strstatus (result));
      gsg_error (buf);
      return;
    }

  if (!win.canvas.gdk_image
      || win.canvas.gdk_image->width != win.params.pixels_per_line
      || win.canvas.gdk_image->height != win.params.lines)
    {
      GdkImageType image_type = GDK_IMAGE_FASTEST;

      if (win.canvas.gdk_image)
	gdk_image_destroy (win.canvas.gdk_image);
#ifdef __alpha__
      /* Some X servers seem to have a problem with shared images that
	 have a width that is not a multiple of 8.  Duh... ;-( */
      if (win.params.pixels_per_line % 8)
	image_type = GDK_IMAGE_NORMAL;
#endif
      win.canvas.gdk_image = gdk_image_new (image_type,
					    win.canvas.preview->window->visual,
					    win.params.pixels_per_line,
					    win.params.lines);
      gtk_widget_set_usize (win.canvas.preview,
			    win.params.pixels_per_line, win.params.lines);
    }

  win.data = win.canvas.gdk_image->mem;
  win.x = 0;

  if (sane_set_io_mode (device_dialog->dev, SANE_TRUE) == SANE_STATUS_GOOD
      && sane_get_select_fd (device_dialog->dev, &fd) == SANE_STATUS_GOOD)
    win.gdk_input_tag = gdk_input_add (fd, GDK_INPUT_READ, input_available, 0);
  else
    input_available (0, -1, 0);
}

static void
play_stop_button (GtkWidget *widget, gpointer client_data, gpointer call_data)
{
  printf ("play_stop_button\n");

  if (win.playing)
    win.playing = FALSE;
  else if (win.gdk_input_tag < 0)
    {
      win.playing = TRUE;
      gtk_label_set (win.play_stop_label, "Stop");

      next_frame ();
    }
}

static void
xcam_exit (void)
{
  gtk_exit (0);
  sane_exit ();
}

int
main (int argc, char **argv)
{
  GtkWidget *menu, *menu_bar, *menu_bar_item, *label, *preview_vbox;
  GtkWidget *hbox, *vbox, *button, *alignment, *frame;
  GdkImageType image_type;
  int i;

  prog_name = strrchr (argv[0], '/');
  if (prog_name)
    ++prog_name;
  else
    prog_name = argv[0];

  sane_init (NULL);

  gdk_set_debug_level (0);
  gdk_set_show_events (0);
  gtk_init (&argc, &argv);

  atexit (xcam_exit);

  win.gdk_input_tag = -1;
  win.shell = gtk_window_new ((char *) prog_name, GTK_WINDOW_TOPLEVEL);
  gtk_widget_show (win.shell);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_set_border_width (vbox, 0);      
  gtk_container_add (win.shell, vbox);
  gtk_widget_show (vbox);

  menu_bar = gtk_menu_bar_new ();

  win.devices.menu = build_device_menu ();

  /* "Files" entry: */
  label = gtk_label_new ("Files");
  gtk_widget_show (label);
  menu_bar_item = gtk_menu_item_new ();
  gtk_container_add (menu_bar, menu_bar_item);
  menu = build_files_menu ();
  gtk_menu_item_set_submenu (menu_bar_item, menu);
  gtk_container_add (menu_bar_item, label);
  gtk_widget_show (menu_bar_item);

  /* "Devices" entry: */
  label = gtk_label_new ("Devices");
  gtk_widget_show (label);
  win.devices.item = gtk_menu_item_new ();
  gtk_container_add (menu_bar, win.devices.item);
  gtk_container_add (win.devices.item, label);
  gtk_menu_item_set_submenu (win.devices.item, win.devices.menu);
  gtk_widget_show (win.devices.item);

  gtk_widget_show (menu_bar);

  gtk_container_add (vbox, menu_bar);

  hbox = gtk_hbox_new (FALSE, 0);
  gtk_container_add (vbox, hbox);
  gtk_widget_show (hbox);

  /* create the device dialog box: */
  win.device_dialog_box = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (win.device_dialog_box);

  /* the preview vbox on the left hand side: */
  preview_vbox = gtk_vbox_new (FALSE, 0);
  gtk_box_pack (hbox, preview_vbox, TRUE, TRUE, 0, GTK_PACK_START);

  frame = gtk_frame_new ("Preview");
  gtk_container_set_border_width (frame, 8);
  gtk_frame_set_type (frame, GTK_SHADOW_ETCHED_IN);
  gtk_container_add (preview_vbox, frame);

  alignment = gtk_alignment_new (0.5, 0.5, 1.0, 1.0);
  gtk_box_pack (preview_vbox, alignment, TRUE, TRUE, 0, GTK_PACK_START);

  win.canvas.preview = gtk_drawing_area_new (320, 200, canvas_events,
					     CANVAS_EVENT_MASK);
  gtk_widget_set_user_data (win.canvas.preview, (gpointer) &win.canvas);
  gtk_container_add (frame, win.canvas.preview);

  gtk_widget_show (win.canvas.preview);
  gtk_widget_show (alignment);
  gtk_widget_show (frame);
  gtk_widget_show (preview_vbox);

  win.canvas.graylevel_cmap = gdk_colormap_get_system ();
  for (i = 0; i < NELEMS(win.canvas.graylevel); ++i)
    {
      GdkColor color;
      color.red = color.green = color.blue =
	i * 0xffff / (NELEMS(win.canvas.graylevel) - 1);
      gdk_color_alloc (win.canvas.graylevel_cmap, &color);
      win.canvas.graylevel[i] = color.pixel;
    }

#if 0
  {
      win.canvas.cube_cmap
	  = gdk_colormap_new (win.canvas.preview->window->visual, 1);
      for (i = 0; i < NELEMS(win.canvas.cube_colors); ++i)
	{
	  win.canvas.cube_colors[i].pixel = i;
	  win.canvas.cube_colors[i].red   = ((i / 30) % 5) * 0xffff / 4;
	  win.canvas.cube_colors[i].green = ((i / 5) % 6)  * 0xffff / 5;
	  win.canvas.cube_colors[i].blue  = ((i % 5))      * 0xffff / 4;
	}
      gdk_colors_store (win.canvas.cube_cmap, win.canvas.cube_colors,
			NELEMS(win.canvas.cube_colors));
      gdk_window_set_colormap (win.shell->window, win.canvas.cube_cmap);
  }
#endif

  gtk_container_add (hbox, win.device_dialog_box);
  if (default_device)
    switch_device (0, (char *) default_device, 0);

  if (device_dialog->dev
      && (sane_get_parameters (device_dialog->dev, &win.params)
	  == SANE_STATUS_GOOD))
    gtk_widget_set_usize (win.canvas.preview,
			  win.params.pixels_per_line, win.params.lines);

  /* buttons at the bottom: */
  button = gtk_push_button_new ();
  gtk_callback_add (gtk_button_get_state (button), play_stop_button, 0);
  win.play_stop_label = gtk_label_new ("Play");
  gtk_container_add (button, win.play_stop_label);
  gtk_container_add (vbox, button);
  gtk_widget_show (win.play_stop_label);
  gtk_widget_show (button);

  gtk_main ();

  sane_exit();
  gtk_exit (0);
  return 0;
}
